//
// This is an example of implementing a multi-valued cursor over a
// simple data structure. This necessarily involves defining a Cursor
// for providing the multiple bindings, and is thus an example of that,
//

import java.util.Hashtable;
import java.util.Observable;
import java.util.Observer;

import aos.jack.jak.logic.ChoicePoint;
import aos.jack.jak.logic.Variable;
import aos.jack.jak.cursor.Cursor;

public class MultiValued extends Observable {

    /**
     * The actual data storage.
     */
    String [] data;

    /**
     * Utility method for adding to data. Note that 1) the data
     * modification is synchronised, and 2) observers are notified
     * after modification.
     * The synchronisation is only required if the data object may be
     * updated by Java threads, which may be clever to assume :-)
     * The notification is added so as to allow the query methods,
     * which return Cursors, to be used in @waitFor statements. Thus,
     * you can write in a plan, something like:
     *
     *     @waitFor( multi.get( "Woop Woop" ) );
     *
     * to wait for the given string to be added to the data.
     */
    synchronized public void add( String d )
    {
	System.err.println( "MultiValued: adding " + d );
	int size = size();
	String [] tmp = new String [ size + 1 ];
	for ( int i=0; i<size; i++ )
	    tmp[ i ] = data[ i ];
	tmp[ size ] = d ;
	data = tmp ;
	setChanged();
	notifyObservers();
    }

    /**
     * Tell size of data.
     */
    public int size()
    {
	return ( data == null )? 0 : data.length;
    }

    /**
     * A simple query example; we don't need the complex query for
     * this.
     */
    public Cursor get( Variable v )
    {
	// Run-time type checking.
	if ( v.getType() != String.class )
	    throw new Error( "`logical String' required." );
	return new DataCursor( v );
    }

    /**
     * The following defines the multi-valued cursor.
     */
    class DataCursor extends Cursor {

	/**
	 * Local data for the cursor; we keep the variable to bind to,
	 * the (current) index into the data table, and choice point
	 * marker into the variable's logical environment.
	 */
	Variable v;
	int index = 0;
	ChoicePoint cp;

	/**
	   Constructor.
	**/
	public DataCursor( Variable v )
	{
	    this.v = v;
	    index = 0;
	    cp = v.getEnv().newChoicePoint();
	}

	/**
	 * Returns true; this method is called by the JACK kernel to
	 * find out whether the Cursor is triggered or not. 
	 */
	public boolean isTriggered()
	{
	    return true;
	}

	/**
	 * Determines the truth-value for this Cursor. When the next()
	 * method returns true, the JACK kernel will call it
	 * repeatedly so as to obtain all successful bindings.
	 * Note that the access is synchronised in order to block data
	 * modifications while we establish the next binding.
	 */
	public boolean next() throws Exception
	{
	    System.err.println( "next: " + this );
	    System.err.println( "index=" + index );

	    if ( cp != null )
		cp.backtrack(); // revert previous binding.

	    if ( data == null )
		return false;

	    synchronized ( MultiValued.this ) {
		System.err.println( "data.length=" + data.length );
		while ( index < data.length ) {
		    if ( v.unify( data [ index++ ] ) )
			return true;
		}
	    }

	    return false;
	}

	/**
	 * Adds an Observer to the Observable tied to this
	 * Cursor. This method is called by the JACK kernel when the
	 * Cursor is used in a @waitFor statement; the Observer
	 * expects to receive a notification as the means of
	 * triggering (asked for above).
	**/
	public void register(Observer o)
	{
	    MultiValued.this.addObserver(o);
	}

	/**
	 * Removes an Observer from the Observable tied to this
	 * Cursor. This method is called by the JACK kernel to clean
	 * up when the Cursor has been used in a @waitFor statement.
	**/
	public boolean unRegister(Observer o)
	{
	    MultiValued.this.deleteObserver(o);
	    return true;
	}
    } /// End of DataCursor

}
